@@ -1,3 +1,4 @@
+# -*- coding: utf-8 -*-
 =begin
 = NArrayMiss Class
 
@@ -35,18 +36,6 @@ To use NArrayMiss class, you need invoking "require 'narray_miss.rb'" in your sc
 
 require 'narray'
 
-class NArray
-  DUMP_SIZE_LIMIT = 20000
-  def self._load(o) to_na(*Marshal::load(o)).ntoh end
-  def _dump(limit)
-    if size <= DUMP_SIZE_LIMIT
-      Marshal::dump([hton.to_s, typecode, *shape])
-    else
-      raise "size of the NArray (#{size}) is too large to dump "+
-            "(limit: #{DUMP_SIZE_LIMIT})"
-    end
-  end
-end
 
 class NArrayMiss
 
@@ -120,7 +109,7 @@ go back to ((<Index>))
      create (({NArrayMiss})) form [((|value|)), ...].
 --- NArrayMiss.to_nam(array [,mask])
      create (({NArrayMiss})) from ((|array|)).
-     ((|array|)) must be (({Array})) or (({NArray})).
+     ((|array|)) must be (({Numeric})) (({Array})) or (({NArray})).
 --- NArrayMiss.to_nam_no_dup(array [,mask])
      convert from ((|array|)) to (({NArrayMiss})).
 
@@ -166,15 +155,15 @@ go back to ((<Index>))
 
     array = arg[0]
     if Numeric===array then array = NArray[array] end
-    if array.class == Array then array = NArray.to_na(array) end
-    if array.class != NArray then
+    if Array===array then array = NArray.to_na(array) end
+    if !array.is_a?(NArray) then
       raise("argument must be Numeric, NArray or Array")
     end
 
     if arg.length==2 then
       mask = arg[1]
-      if Numeric===mask then mask = [mask] end
-      if mask.class == Array then
+      if Numeric===mask then mask = array.ne(mask) end
+      if Array===mask then
 	mask = NArray.to_na(mask).ne(0)
       end
       if mask.class == FalseClass then
@@ -183,7 +172,7 @@ go back to ((<Index>))
       if mask.class == TrueClass then
 	mask = NArray.byte(*array.shape).fill(1)
       end
-      if !(mask.class==NArray && mask.typecode==BYTE) then
+      if !(NArray===mask && mask.typecode==BYTE) then
 	  raise("mask must be Numeric, Array, true, false or NArray(byte)")
       end
       if mask.length!=array.length
@@ -195,7 +184,7 @@ go back to ((<Index>))
     __new__(array,mask)
   end
   def self.to_nam(*arg)
-    if !(Numeric===arg[0]) && arg[0].class!=Array && arg[0].class!=NArray
+    if !(Numeric===arg[0]) && !(Array===arg[0]) && !arg[0].is_a?(NArray)
       raise "first argument must be Numeric, NArray or Array"
     end
     arg[0] = arg[0].dup if !(Numeric===arg[0])
@@ -272,11 +261,20 @@ go back to ((<Index>))
 =end
 
   def [](*arg)
-    obj = @array[*arg]
-    if Numeric===obj
-      return obj
+    if arg[0].class == NArrayMiss && arg[0].typecode == BYTE
+      obj = @array[arg[0].to_na(0)]
+      if Numeric===obj
+        return obj
+      else
+        return NArrayMiss.to_nam_no_dup(obj)
+      end
     else
-      return NArrayMiss.to_nam_no_dup(obj,@mask[*arg])
+      obj = @array[*arg]
+      if Numeric===obj
+        return obj
+      else
+        return NArrayMiss.to_nam_no_dup(obj,@mask[*arg])
+      end
     end
   end
   def slice(*arg)
@@ -307,12 +305,20 @@ go back to ((<Index>))
   end
 
   def []=(*arg)
-    self.set_without_validation(*arg)
-    if arg[-1].class != NArrayMiss && arg[-1] then
-      if arg.length==1 then
-        @mask=1
-      else
-        @mask[*arg[0..-2]] = 1
+    if arg.length == 2 && arg[0].class == NArrayMiss && arg[0].typecode == BYTE
+      idx = arg[0].to_na(0)
+      self.set_without_validation(idx,arg[-1])
+      if arg[-1].class != NArrayMiss && arg[-1] then
+        @mask[idx] = 1
+      end
+    else
+      self.set_without_validation(*arg)
+      if arg[-1].class != NArrayMiss && arg[-1] then
+        if arg.length==1 then
+          @mask=1
+        else
+          @mask[*arg[0..-2]] = 1
+        end
       end
     end
     return self
@@ -337,7 +343,7 @@ go back to ((<Index>))
      set normally distributed random values with mean=0, dispersion=1 (Box-Muller)
 =end
 
-  for operator in ["indgen","fill","random"]
+  for operator in ["indgen","fill","random","randomn"]
     eval(<<-EOL,nil,__FILE__,__LINE__+1)
     def #{operator}(*arg)
       obj = self.dup
@@ -351,11 +357,6 @@ go back to ((<Index>))
     end
     EOL
   end
-  def randomn
-    obj = self.dup
-    obj[@mask] = @array[@mask].randomn
-    obj
-  end
 
 
 =begin
@@ -381,37 +382,47 @@ go back to ((<Index>))
     array[@mask] = -@array[@mask]
     NArrayMiss.to_nam_no_dup(array, @mask.dup)
   end
-  for operator in ["+","-","*","/","%","**"]
-    dummy = {"+"=>0,"-"=>0,"*"=>1,"/"=>1,"%"=>1,"**"=>1}[operator]
-    eval(<<-EOL,nil,__FILE__,__LINE__+1)
-    def #{operator}(arg)
-      if !arg then
-	@mask = 0
-	return self
-      else
-        term1,term2,mask,flag = routine1(arg,#{dummy})
-        result = term1 #{operator} term2
-        routine2(result,mask,flag)
-      end
-    end
-    EOL
+  def +(arg)
+    binary_operation(arg, 0){|t1, t2| t1 + t2}
   end
+  def -(arg)
+    binary_operation(arg, 0){|t1, t2| t1 - t2}
+  end
+  def *(arg)
+    binary_operation(arg, 1){|t1, t2| t1 * t2}
+  end
+  def /(arg)
+    binary_operation(arg, 1){|t1, t2| t1 / t2}
+  end
+  def %(arg)
+    binary_operation(arg, 1){|t1, t2| t1 % t2}
+  end
+  def **(arg)
+    binary_operation(arg, 1){|t1, t2| t1 ** t2}
+  end
+
   def abs
     array = @array.dup
     array[@mask] = @array[@mask].abs
     NArrayMiss.to_nam_no_dup(array, @mask.dup)
   end
 
-  for operator in ["add!","sbt!","mul!","div!","mod!"]
-    dummy = {"add!"=>0,"sbt!"=>0,"mul!"=>1,"div!"=>1,"mod!"=>1}[operator]
-    eval(<<-EOL,nil,__FILE__,__LINE__+1)
-    def #{operator}(arg)
-      term1,term2,mask,flag = routine1(arg,#{dummy})
-      result = term1.#{operator} term2
-      routine2(result,mask,flag)
-    end
-    EOL
+  def add!(arg)
+    binary_operation(arg, 0){|t1, t2| t1.add!(t2)}
   end
+  def sbt!(arg)
+    binary_operation(arg, 0){|t1, t2| t1.sbt!(t2)}
+  end
+  def mul!(arg)
+    binary_operation(arg, 1){|t1, t2| t1.mul!(t2)}
+  end
+  def div!(arg)
+    binary_operation(arg, 1){|t1, t2| t1.div!(t2)}
+  end
+  def mod!(arg)
+    binary_operation(arg, 1){|t1, t2| t1.mod!(t2)}
+  end
+
   def mul_add(*arg)
     if arg.length==1 then
       return (self*arg[0]).sum
@@ -432,15 +443,14 @@ go back to ((<Index>))
   def ~@
     NArrayMiss.to_nam_to_dup(~@array, @mask.dup)
   end
-  for operator in ["&","|","^"]
-    dummy = {"&"=>1,"|"=>0,"^"=>1}[operator]
-    eval(<<-EOL,nil,__FILE__,__LINE__+1)
-    def #{operator}(arg)
-      term1,term2,mask,flag = routine1(arg,#{dummy})
-      result = term1 #{operator} term2
-      routine2(result,mask,flag)
-    end
-    EOL
+  def &(arg)
+    binary_operation(arg, 1){|t1, t2| t1 & t2}
+  end
+  def |(arg)
+    binary_operation(arg, 0){|t1, t2| t1 | t2}
+  end
+  def ^(arg)
+    binary_operation(arg, 1){|t1, t2| t1 ^ t2}
   end
 
 
@@ -462,46 +472,46 @@ go back to ((<Index>))
 --- NArrayMiss#not(other)
 =end
 
-  for operator in ["eq","ne","gt","ge","lt","le"]
-    eval(<<-EOL,nil,__FILE__,__LINE__+1)
-    def #{operator}(arg)
-      term1,term2,mask,flag = routine1(arg,0)
-      result = term1.#{operator} term2
-      routine2(result,mask,flag)
-    end
-    EOL
+  def eq(arg)
+    binary_operation(arg, 0){|t1, t2| t1.eq t2}
   end
-  for operator in [">",">=","<","<="]
-    eval(<<-EOL,nil,__FILE__,__LINE__+1)
-    def #{operator}(arg)
-      term1,term2,mask,flag = routine1(arg,0)
-      result = term1 #{operator} term2
-      routine2(result,mask,flag)
-    end
-    EOL
+  def ne(arg)
+    binary_operation(arg, 0){|t1, t2| t1.ne t2}
+  end
+  def gt(arg)
+    binary_operation(arg, 0){|t1, t2| t1.gt t2}
+  end
+  def ge(arg)
+    binary_operation(arg, 0){|t1, t2| t1.ge t2}
+  end
+  def lt(arg)
+    binary_operation(arg, 0){|t1, t2| t1.lt t2}
+  end
+  def le(arg)
+    binary_operation(arg, 0){|t1, t2| t1.le t2}
   end
 
-  for operator in ["and","or","xor"]
-    dummy = {"and"=>1,"or"=>0,"xor"=>1}[operator]
-    eval(<<-EOL,nil,__FILE__,__LINE__+1)
-    def #{operator}(arg)
-      term1,term2,mask,flag = routine1(arg,#{dummy})
-      result = term1.#{operator} term2
-      routine2(result,mask,flag)
-    end
-    EOL
+  def and(arg)
+    binary_operation(arg, 1){|t1, t2| t1.and t2}
   end
+  def or(arg)
+    binary_operation(arg, 0){|t1, t2| t1.or t2}
+  end
+  def xor(arg)
+    binary_operation(arg, 1){|t1, t2| t1.xor t2}
+  end
+
   def not
     NArrayMiss.to_nam_no_dup(@array.not, @mask.dup)
   end
 
-#  def ==(arg)
-#    if art.kind_of?(NArrayMiss) then
-#      @array==arg.get_array! && @mask==arg.get_mask!
-#    else
-#      false
-#    end
-#  end
+  def ==(arg)
+    if arg.kind_of?(NArrayMiss) then
+      @mask==arg.get_mask! && @array[@mask]==arg.get_array![@mask]
+    else
+      false
+    end
+  end
 
 
 =begin
@@ -517,12 +527,21 @@ go back to ((<Index>))
 --- NArrayMiss#max(dim, ...)
      return maximum in specified dimensions.
      Elements at which the number of valid elements in the dimension is less than ((|i|)) is invalid.
+--- NArrayMiss#median(dim, ...)
+     return median in specified dimensions.
+     Elements at which the number of valid elements in the dimension is less than ((|i|)) is invalid.
 --- NArrayMiss#mean(dim, ...)
      return mean of elements in specified dimensions.
      Elements at which the number of elements for then mean is less than ((|i|)) is invalid.
 --- NArrayMiss#stddev(dim, ...)
      return standard deviation of elements in specified dimensions.
      Elements at which the number of elements for calculation the standard deviation is less than ((|i|)) is invalid.
+--- NArrayMiss#rms(dim, ...)
+     return root mean square of elements in specified dimensions.
+     Elements at which the number of elements for calculation the RMS is less than ((|i|)) is invalid.
+--- NArrayMiss#rmsdev(dim, ...)
+     return root mean square deviation of elements in specified dimensions.
+     Elements at which the number of elements for calculation the RMS deviation is less than ((|i|)) is invalid.
 =end
 
   def accum(*arg)
@@ -536,123 +555,101 @@ go back to ((<Index>))
     end
   end
 
-  for operator in ["sum","min","max"]
-    str = {"sum"=>"0","min"=>"array.max","max"=>"array.min"}[operator]
-    eval(<<-EOL,nil,__FILE__,__LINE__+1)
-    def #{operator}(*arg)
-      if @mask.count_true == 0 then
-        return nil
-      end
-      min_count=1
-      options = ["min_count"]
-      if arg.length!=0 && arg[-1].class==Hash then
-	option = arg[-1]
-	arg = arg[0..-2]
-	option.each_key{|key|
-	  if !options.index(key) then
-	    raise(ArgumentError,key+" option is not exist")
-	  end
-	}
-	min_count = option["min_count"]
-      end
-      mask = @mask.to_type(NArray::INT)
-      array = @array.dup
-      array[@mask.not] = #{str}
-      mask = mask.sum(*arg)
-      if NArray===mask then
-	if mask.ge(min_count).count_true == 0 then
-	  return nil
-        else
-          return NArrayMiss.to_nam_no_dup(array.#{operator}(*arg),
-                                          mask.ge(min_count))
-        end
-      else
-        if mask < min_count
-          return nil
-        else
-          return array.#{operator}
-        end
-      end
-    end
-    EOL
-  end
-  def mean(*arg)
-    if @mask.count_true == 0 then
-      return nil
-    end
-    min_count = 1
-    options=["min_count"]
-    if arg.length!=0 && arg[-1].class==Hash then
-      option = arg[-1]
-      arg2=arg[0..-2]
-      option.each_key{|key|
-	if !options.index(key) then
-	  raise(ArgumentError,key+" option is not exist")
-	end
-      }
-      min_count = option["min_count"]
-    else
-      arg2=arg
-    end
-    count = @mask.to_type(NArray::INT).sum(*arg2)
-    if count.class == NArray then
-      count = NArrayMiss.to_nam(count,count.ge(min_count))
-      if count.ge(min_count).count_true == 0
-        return nil
-      else
-        return self.sum(*arg2)/count
-      end
-    else
-      if count < min_count then
-        return nil
-      else
-	return self.sum(*arg2)/count
-      end
+  def sum(*dims)
+    min_count = NArrayMiss.check_options(dims, 1)
+    # 欠損値に 0 を入れて普通に sum する
+    ary0 = @array.dup
+    ary0[@mask.not] = 0
+    NArrayMiss.reduction(@mask, min_count, dims, false, typecode) do
+      ary0.sum(*dims)
     end
   end
-  def stddev(*arg)
-    if @mask.count_true == 0 then
-      return nil
+  def min(*dims)
+    min_count = NArrayMiss.check_options(dims, 1)
+    # 欠損値に最大値を入れて普通に min する
+    # byte,sint,int,sfloat,float の MAX の値を入れるように変更すべき
+    ary0 = @array.dup
+    ary0[@mask.not] = @array.max
+    NArrayMiss.reduction(@mask, min_count, dims, false, typecode) do
+      ary0.min(*dims)
     end
-    min_count=2
-    options=["min_count"]
-    if arg.length!=0 && arg[-1].class==Hash then
-      option = arg[-1]
-      arg = arg[0..-2]
-      option.each_key{|key|
-	if !options.index(key) then
-	  raise(ArgumentError,key+" option is not exist")
-	end
-      }
-      min_count = option["min_count"].to_i
-      if min_count<2
-        raise(ArgumentError, "min_count must be >= 2")
-      end
+  end
+  def max(*dims)
+    min_count = NArrayMiss.check_options(dims, 1)
+    # 欠損値に最小値を入れて普通に max する
+    # byte,sint,int,sfloat,float の MIN の値を入れるように変更すべき
+    ary0 = @array.dup
+    ary0[@mask.not] = @array.min
+    NArrayMiss.reduction(@mask, min_count, dims, false, typecode) do
+      ary0.max(*dims)
     end
-    count = @mask.to_type(NArray::INT).sum(*arg)
-    count2 = @mask.to_type(NArray::INT).accum(*arg)
-    if count.class==NArray then
-      count = NArrayMiss.to_nam_no_dup(count,count.ge(min_count))
-      if count.get_mask!.count_true == 0
-        return nil
-      end
-    else
-      if count < min_count then
-	return nil
-      end
-    end
-    if self.integer? then
-      a = self.to_f
-    else
-      a = self
-    end
-    var = ( (a-a.accum(*arg)/count2)**2 ).sum(*arg)/(count-1)
-    obj = NMMath::sqrt(var)
-    return obj
   end
 
-#  def median(*arg)
-#  end
+  def mean(*dims)
+    min_count = NArrayMiss.check_options(dims, 1)
+    # 整数型の場合は浮動小数型へ変換
+    ary0 = self.integer? ? self.to_type(NArray::DFLOAT) : self
+    NArrayMiss.reduction(@mask, min_count, dims, true, typecode) do |count_sum, count_accum|
+      var = ary0.sum(*dims)/count_sum
+    end
+  end
+  def stddev(*dims)
+    min_count = NArrayMiss.check_options(dims, 2)
+    # 整数型の場合は浮動小数型へ変換
+    ary0 = self.integer? ? self.to_type(NArray::DFLOAT) : self
+    NArrayMiss.reduction(@mask, min_count, dims, true, typecode) do |count_sum, count_accum|
+      ary0 = ary0 - ary0.accum(*dims)/count_accum
+      ary0 = ary0.abs if ary0.complex?
+      ary0 = (ary0**2).sum(*dims) / (count_sum-1)
+      NMMath.sqrt(ary0)
+    end
+  end
+  def rms(*dims)
+    min_count = NArrayMiss.check_options(dims, 1)
+    # 整数型の場合は浮動小数型へ変換
+    ary0 = self.integer? ? self.to_type(NArray::DFLOAT) : self
+    NArrayMiss.reduction(@mask, min_count, dims, true, typecode) do |count_sum, count_accum|
+      ary0 = ary0.abs if ary0.complex?
+      ary0 = (ary0**2).sum(*dims) / count_sum
+      NMMath.sqrt(ary0)
+    end
+  end
+  def rmsdev(*dims)
+    min_count = NArrayMiss.check_options(dims, 1)
+    # 整数型の場合は浮動小数型へ変換
+    ary0 = self.integer? ? self.to_type(NArray::DFLOAT) : self
+    NArrayMiss.reduction(@mask, min_count, dims, true, typecode) do |count_sum, count_accum|
+      ary0 = ary0 - ary0.accum(*dims)/count_accum
+      ary0 = ary0.abs if ary0.complex?
+      ary0 = (ary0**2).sum(*dims) / count_sum
+      NMMath.sqrt(ary0)
+    end
+  end
+
+  def median(*arg)
+    if arg.length==0 then
+      return @array[@mask].median
+    else
+      nshape = NArray.to_na(@array.shape)
+      nshape[arg]=1
+      nslice = nshape[nshape.ne(1).where]
+      index = NArray.object(@mask.rank)
+      index[nshape.eq(1).where] = true
+      obj = NArrayMiss.new(@array.typecode,*nslice.to_a)
+      total = 1
+      nslice.each{|n| total *= n}
+      for i in 0...total
+        index[nshape.ne(1).where] = pos(i,nslice)
+        mask = NArray.byte(*@array.shape).fill(0)
+        mask[*index] = 1
+        mask = @mask&mask
+        if mask.count_true != 0 then
+          obj[*pos(i,nslice)] = @array[mask].median
+        end
+      end
+      return obj
+    end
+  end
 
 
 =begin
@@ -671,20 +668,20 @@ go back to ((<Index>))
 	obj[@mask] = @array[@mask].#{operator}
         return obj
       else
-	shape=@array.shape
-	shape[*arg]=1
-	nshape = NArray.to_na(shape)
-	slice = shape[*nshape.ne(1).where]
-	nslice = NArray.to_na(slice)
-	index = NArray.object(@mask.rank)
-	index[*nshape.eq(1).where] = true
-	obj = NArrayMiss.new(@array.typecode,*@array.shape)
-	for i in 0..nslice.sum-1
-	  index[*nshape.ne(1).where] = pos(i,nslice)
-	  mask = NArray.byte(*@array.shape).fill(0)
-	  mask[*index.to_a] = 1
-	  mask = @mask&mask
-	  if mask.count_true != 0 then
+        nshape = NArray.to_na(@array.shape)
+        nshape[arg]=1
+        nslice = nshape[nshape.ne(1).where]
+        index = NArray.object(@mask.rank)
+        index[nshape.eq(1).where] = true
+        obj = NArrayMiss.new(@array.typecode,*@array.shape)
+        total = 1
+        nslice.each{|n| total *= n}
+        for i in 0...total
+          index[nshape.ne(1).where] = pos(i,nslice)
+          mask = NArray.byte(*@array.shape).fill(0)
+          mask[*index] = 1
+          mask = @mask&mask
+          if mask.count_true != 0 then
 	    obj[mask] = @array[mask].#{operator}
 	  end
 	end
@@ -781,13 +778,22 @@ go back to ((<Index>))
      create (({NArrayMiss})) of object whose elements are processed (({to_s}))
 =end
 
-  for operator in ["floor","ceil","round","to_i","to_f"]
-    eval(<<-EOL,nil,__FILE__,__LINE__+1)
-    def #{operator}
-      NArrayMiss.to_nam_no_dup(@array.#{operator}, @mask.dup)
-    end
-    EOL
+  def floor
+    NArrayMiss.to_nam_no_dup(@array.floor, @mask.dup)
   end
+  def ceil
+    NArrayMiss.to_nam_no_dup(@array.ceil, @mask.dup)
+  end
+  def round
+    NArrayMiss.to_nam_no_dup(@array.round, @mask.dup)
+  end
+  def to_i
+    NArrayMiss.to_nam_no_dup(@array.to_i, @mask.dup)
+  end
+  def to_f
+    NArrayMiss.to_nam_no_dup(@array.to_f, @mask.dup)
+  end
+
   def to_type(typecode)
     NArrayMiss.to_nam_no_dup(@array.to_type(typecode), @mask.dup)
   end
@@ -967,11 +973,11 @@ go back to ((<Index>))
     obj
   end
   def hton
-    NArray.to_nam(@array.hton,@mask.hton)
+    NArrayMiss.to_nam(@array.hton,@mask.hton)
   end
   alias :ntoh :hton
   def htov
-    NArray.to_nam(@array.htov,@mask.htov)
+    NArrayMiss.to_nam(@array.htov,@mask.htov)
   end
   alias :vtoh :htov
 
@@ -1023,10 +1029,12 @@ go back to ((<Index>))
 
   def set_valid(*pos)
     @mask[*pos] = 1
+    self
   end
   alias validation set_valid
   def set_invalid(*pos)
     @mask[*pos] = 0
+    self
   end
   alias invalidation set_invalid
   def all_valid
@@ -1173,20 +1181,20 @@ go back to ((<Index>))
     sep = ", "
     const = Hash.new
     NArray.constants.each{|c| const[NArray.const_get(c)] = c}
-    str_ret = "NArrayMiss."+const[typecode].downcase+"("+shape.join(",")+"):"
+    str_ret = "NArrayMiss."+const[typecode].to_s.downcase+"("+shape.join(",")+"):"
     if rank == 0 then
-      str_ret += " []"
+      str_ret << " []"
       return str_ret
     else
-       str_ret += "\n"
+       str_ret << "\n"
     end
     str = ""
     index = Array.new(rank,0)
     index[0] = true
     i = 1
-    (rank-1).times{ str_ret += "[ " }
+    (rank-1).times{ str_ret << "[ " }
     while(true)
-      i.times{ str_ret += "[ " }
+      i.times{ str_ret << "[ " }
 
       str = @array[*index].inspect
       ary = str[str.index("[")+1..str.index("]")-1].strip.split(/\s*,\s*/)
@@ -1200,30 +1208,30 @@ go back to ((<Index>))
         ary.pop
         ary[-1] = "..."
       end
-      str_ret += ary.join(", ")
+      str_ret << ary.join(", ")
       i = 1
       while (i<rank)
 	if index[i]<shape[i]-1 then
-	  str_ret += " ]"+sep+"\n"
+	  str_ret << " ]" << sep << "\n"
 	  count_line += 1
 	  index[i] += 1
 	  break
 	else
-	  str_ret += " ]"
+	  str_ret << " ]"
 	  index[i] = 0
 	  i += 1
 	end
       end
 
       if i>=rank then
-	str_ret += " ]"
+	str_ret << " ]"
 	return str_ret
       elsif count_line>=max_line then
-	str_ret += " ..."
+	str_ret << " ..."
 	return str_ret
       end
 
-      (rank-i).times{ print("  ") }
+      (rank-i).times{ str_ret << "  " }
     end
     return str_ret
   end
@@ -1256,54 +1264,98 @@ go back to ((<Index>))
     result[0] = m
     result
   end
-  def routine1(arg,dummy)
+  def binary_operation(arg, dummy)
+    # arg: 第2項目のオブジェクト
+    # dummy: 演算を行っても結果に影響を与えない特別な値。(欠損部分に代入する)
     flag=true
-    if Numeric===arg then
+    case arg
+    when Numeric
       term1 = @array
       term2 = arg
-      mask = true
-    elsif arg.class == Array then
-      term1 = @array
+      mask = @mask
+    when Array, NArray
+      term1 = @array.dup
+      term1[@mask.not] = dummy # 欠損部分に dummy を代入
+      term2 = arg.kind_of?(NArray) ? arg : NArray.to_na(arg) # Array -> NArray
+      mask = NArray.byte(*term2.shape).fill(1) # 2項目は欠損無し
+      mask = @mask & mask
+    when NArrayMiss
+      term1 = @array.dup
       term1[@mask.not] = dummy
-      term2 = NArray.to_na(arg)
-      mask = NArray.byte(*term2.shape).fill(1)
-    elsif arg.class == NArray then
-      term1 = @array
-      term1[@mask.not] = dummy
-      term2 = arg
-      mask = NArray.byte(*term2.shape).fill(1)
-    elsif arg.class == NArrayMiss then
-      mask = arg.get_mask
-      term1 = @array
-      term1[@mask.not] = dummy
-      term2 = arg.to_na
-      term2[arg.get_mask!.not] = dummy
+      mask = arg.get_mask!
+      term2 = arg.get_array
+      term2[mask.not] = dummy
+      mask = @mask & mask
     else
-      term1,term2 = arg.coerce(self)
-      flag=false
+      term1, term2 = arg.coerce(self)
+      # 演算を arg のクラスに任せるため、yield の結果をそのまま返す
+      flag = false
     end
-    [term1,term2,mask,flag]
+    result = yield(term1, term2)
+    result = NArrayMiss.to_nam_no_dup(result, mask) if flag
+    result
   end
-  def routine2(result,mask,flag)
-    if flag then
-      obj = NArrayMiss.to_nam_no_dup(result)
-      if mask==true then
-	obj.set_mask(@mask)
+
+
+  def self.reduction(mask, min_count, dims, flag, typecode)
+    # flag: リダクションを行う次元方向の有効な値の個数で、割り算を行うかどうかのフラグ
+    count_sum = mask.to_type(NArray::LINT).sum(*dims)
+    # 返り値が配列か、スカラーかによって分岐
+    if count_sum.kind_of?(NArray)
+      mask = count_sum.ge(min_count)
+      # すべての要素が欠損値にならないかチェック
+      if mask.any?
+        count_accum = NArray.ref(count_sum)
+        dims.collect{|d|d<0 ? d+mask.rank : d}.sort.each do |d|
+          count_accum.newdim!(d)
+        end
+        # 割り算を行う場合は、先に count_sum を NArrayMiss 化
+        #   yield の戻り値は NArrayMiss
+        # 割り算を行わない場合は、後で NArrayMiss 化
+        #   yield の戻り値は NArray
+        count_sum = NArrayMiss.to_nam_no_dup(count_sum,mask) if flag
+        ary = yield(count_sum, count_accum)
+        ary = NArrayMiss.to_nam_no_dup(ary, mask) unless flag
       else
-	mask = @mask+mask
-	obj.set_mask(mask.eq(2))
-      end
-      obj.set_without_validation(@mask.not,@array[@mask.not])
-      if Numeric===obj && obj.get_mask![0]==1 then
-	return obj.get_array[0]
-      else
-	return obj
+        # すべての要素が欠損値の NArrayMiss を返す
+        na = NArray.new(typecode, *mask.shape)
+        ary = NArrayMiss.to_nam_no_dup(na, false)
       end
     else
-      result
+      # 有効な要素数があるかチェック
+      if count_sum >= min_count
+        count_accum = NArray.int(*([1]*mask.rank)).fill!(count_sum)
+        ary = yield(count_sum, count_accum)
+      else
+        # 有効な要素数が足りない場合は nil を返す
+        return nil
+      end
     end
+    return ary
   end
+
     
+  # 引数にオプション (Hash) が指定されているかチェックし、
+  # されている場合は、オプションを取得し戻り値として返す。
+  # 現時点では、オプションは "min_count" のみ
+  def self.check_options(arg, default_mincount)
+    min_count = default_mincount
+    options = %w(min_count)
+    if arg.length!=0 && arg[-1].kind_of?(Hash)
+      option = arg.pop
+      option.each_key{|key|
+	if !options.index(key) then
+	  raise(ArgumentError,key+" option is not exist")
+	end
+      }
+      min_count = option["min_count"] || default_mincount
+      min_count = min_count.to_i
+      if min_count < default_mincount
+        raise(ArgumentError, "min_count must be >= #{default_mincount}")
+      end
+    end
+    return min_count
+  end
 end
 
 
@@ -1313,21 +1365,27 @@ module NMMath
 
   func1 =  ["sqrt","exp","log","log10","log2",
             "sin","cos","tan","sinh","cosh","tanh",
-            "asin","acos","atan","asinh","acosh","atanh"]
+            "asin","acos","atan","asinh","acosh","atanh",
+            "csc", "sec", "cot", "csch", "sech", "coth",
+            "acsc", "asec", "acot", "acsch", "asech", "acoth",
+           ]
   func2 = ["atan2"]
 
   for operator in func1
     eval <<-EOL,nil,__FILE__,__LINE__+1
     def #{operator}(x)
-      if Numeric===x || x.class==Array || x.class==NArray then
+      case x
+      when Numeric, Array, NArray
 	NMath::#{operator}(x)
-      elsif x.class == NArrayMiss then
+      when NArrayMiss
 	obj = NArrayMiss.new(x.typecode,*x.shape)
 	mask = x.get_mask!
 	obj[mask] = NMath::#{operator}(x.get_array![mask])
 	obj[mask.not] = x[mask.not]
 	obj.set_mask(mask)
         obj
+      else
+        raise ArgumentError, "argument is invalid class: \#{x.class}"
       end
     end
     module_function :#{operator}
@@ -1337,17 +1395,24 @@ module NMMath
   for operator in func2
     eval <<-EOL,nil,__FILE__,__LINE__+1
     def #{operator}(x,y)
-      if Numeric===x || x.class==Array || x.class==NArray then
+      obj = nil
+      case x
+      when Numeric, Array, NArray
 	mask1 = nil
-      elsif x.class == NArrayMiss then
+      when NArrayMiss
 	obj = NArrayMiss.new(x.typecode,*x.shape)
 	mask1 = x.get_mask!
+      else
+        raise ArgumentError, "argument is invalid class"
       end
-      if Numeric===y || y.class==Array || y.class==NArray then
+      case y
+      when Numeric, Array, NArray
 	mask2 = nil
-      elsif y.class == NArrayMiss then
-	obj = NArrayMiss.new(y.typecode,*y.shape)
+      when NArrayMiss
+	obj ||= NArrayMiss.new(y.typecode,*y.shape)
 	mask2 = y.get_mask!
+      else
+        raise ArgumentError, "argument is invalid class"
       end
       if mask2.nil? then
 	if mask1.nil? then
@@ -1366,7 +1431,6 @@ module NMMath
           return obj
         else
           obj[mask1&mask2] = NMath::#{operator}(x.get_array![mask1],y.get_array![mask2])
-          obj[(mask1&mask2).not] = y[(mask1&mask2).not]
           obj[(mask1&mask2).not] = x[(mask1&mask2).not]
           return obj
         end
@@ -1376,14 +1440,69 @@ module NMMath
     EOL
   end
 
-  for operator in func1+func2
+  methods = Hash.new
+  methods["covariance"] = {:min_count => 2, :post => <<-EOL}
+    mean0 = ary0.accum(*dims) / count_accum
+    mean1 = ary1.accum(*dims) / count_accum
+    ary = ((ary0-mean0)*(ary1-mean1)).sum(*dims)
+    ary / (count_sum-1)
+  EOL
+  methods.each do |name, hash|
+    eval <<-EOL,nil,__FILE__,__LINE__+1
+      def #{name}(ary0,ary1,*dims)
+        min_count = NArrayMiss.check_options(dims, 2)
+        case ary0
+        when Numeric, Array, NArray
+          mask0 = nil
+        when NArrayMiss
+          mask0 = ary0.get_mask!
+        else
+          raise ArgumentError, "argument is invalid class"
+        end
+        case ary1
+        when Numeric, Array, NArray
+          mask1 = nil
+        when NArrayMiss
+          mask1 = ary1.get_mask!
+        else
+          raise ArgumentError, "argument is invalid class"
+        end
+        if mask1.nil? then
+          if mask0.nil? then
+            return NMath.#{name}(ary0, ary1, *dims)
+          else
+            ary1 = NArrayMiss.to_nam_no_dup(ary1, true)
+          end
+        else
+          if mask0.nil?
+            ary0 = NArrayMiss.to_nam_no_dup(ary0, true)
+          end
+        end
+        if ary1.shape != ary1.shape
+          raise "shape is different"
+        end
+        ary0 = ary0.to_type(NArray::DFLOAT)
+        ary1 = ary1.to_type(NArray::DFLOAT)
+        mask = mask0.nil? ? mask1 : mask1.nil? ? mask0 : mask0&mask1
+        ary0.set_mask(mask)
+        ary1.set_mask(mask)
+        NArrayMiss.reduction(mask, min_count, dims, true, NArray::DFLOAT) do |count_sum, count_accum|
+          #{hash[:post]}
+        end
+      end
+      module_function :#{name}
+    EOL
+  end
+
+  for operator in func1+func2+["covariance"]
     eval <<-EOL,nil,__FILE__,__LINE__+1
     def #{operator}(*x)
       x = [self]+x if NArrayMiss===self
-      NMMath::#{operator}(*x)
+      NMMath.#{operator}(*x)
     end
     EOL
-  end
+ end
+
 end
 
 class NArrayMiss
